#include <cstdint>
#include <iomanip>
#include <iostream>
#include <vector>

uint32_t doubleFloatBitwise(uint32_t in);
int popCount(uint32_t in);
uint32_t doubleFloatAsm(uint32_t in);
int popCountAsm(uint32_t in);
bool testSingleDouble(uint32_t in);
bool testSinglePopCount(uint32_t in);
bool testAllDouble();
bool testAllPop();
bool testSpecifiedPop(std::vector<uint32_t> in);
bool testSpecifiedDouble(std::vector<float> in);

uint32_t doubleFloatBitwise(uint32_t in) {
  // Extract the parts
  uint32_t sign = in >> 31;
  uint32_t exponent = (in >> 23) & 0xFF;
  uint32_t mantissa = in & uint32_t{0x7FFFFF};

  if (exponent == 0xFF) {
    // Special value

    // Leave infty alone
    if (mantissa == 0)
      return in;

    // NaN get this bit flipped
    mantissa |= 0x400000;
  } else if (exponent != 0) {
    // Normalized
    ++exponent;
    exponent &= 0xFF;

    if (exponent == 0xFF) {
      // Overflowed to positive infinity
      mantissa = 0;
    }
  } else {
    // Subnormal

    // If this overflows it goes into the mantissa as desired.
    mantissa <<= 1;
  }

  // Reassemble the bits
  in = (sign << 31) | (exponent << 23) | mantissa;

  return in;
}

// Population count - Count the number of bits set in the input
int popCount(uint32_t in) {
  int pop;

  for (pop = 0; in != 0; ++pop)
    in &= in - 1;

  return pop;
}


bool testSingleDouble(uint32_t in) {
  bool match = doubleFloatBitwise(in) == doubleFloatAsm(in);

  if (!match) {
    // Store the current format status of std::out
    std::ios state(nullptr);
    state.copyfmt(std::cout);

    std::cerr << "Fail match for doubling 0x" << std::hex << std::setfill('0')
              << std::setw(8) << in << "\n";
    // Restore the format status of std::out
    std::cout.copyfmt(state);
  }

  return match;
}

bool testSinglePopCount(uint32_t in) {
  bool match = popCount(in) == popCountAsm(in);

  if (!match) {
    // Store the current format status of std::out
    std::ios state(nullptr);
    state.copyfmt(std::cout);

    std::cerr << "Fail match for popcount 0x" << std::hex << std::setfill('0')
              << std::setw(8) << in << "\n";
    // Restore the format status of std::out
    std::cout.copyfmt(state);
  }

  return match;
}

bool testAllDouble() {
  std::cerr << "Starting testAllDouble\n";

  bool pass = true;
  for (uint64_t i = 0; i < (uint64_t{1} << 33); ++i) {
    if (!testSingleDouble(uint32_t(i)))
      pass = false;
  }

  std::cerr << "Finished testAllDouble. ";
  if (pass)
    std::cerr << "Passed.\n";
  else
    std::cerr << "Failed.\n";

  return pass;
}

bool testAllPop() {
  std::cerr << "Starting testAllPop\n";

  bool pass = true;
  for (uint64_t i = 0; i < (uint64_t{1} << 33); ++i) {
    if (!testSinglePopCount(uint32_t(i)))
      pass = false;
  }

  std::cerr << "Finished testAllPop. ";
  if (pass)
    std::cerr << "Passed.\n";
  else
    std::cerr << "Failed.\n";

  return pass;
}

bool testSpecifiedPop(std::vector<uint32_t> in) {
  std::cerr << "Starting testSpecifiedPop on {";
  for (const auto i : in)
    std::cerr << i << ", ";
  std::cerr << "}\n";

  bool pass = true;

  for (const auto i : in) {
    if (!testSinglePopCount(i))
      pass = false;
  }

  std::cerr << "Finished testSpecifiedPop. ";
  if (pass)
    std::cerr << "Passed.\n";
  else
    std::cerr << "Failed.\n";

  return pass;
}

bool testSpecifiedDouble(std::vector<float> in) {
  std::cerr << "Starting testSpecifiedPop on {";
  for (const auto i : in)
    std::cerr << i << ", ";
  std::cerr << "}\n";

  bool pass = true;

  for (const auto i : in) {
    if (!testSinglePopCount(i))
      pass = false;
  }

  std::cerr << "Finished testSpecifiedPop. ";
  if (pass)
    std::cerr << "Passed.\n";
  else
    std::cerr << "Failed.\n";

  return pass;
}

int main() {
  testSpecifiedPop({{0, 1, 2, 3}});
  testSpecifiedDouble({{-1, -0.5, 0, 0.5, 1}});
  testAllDouble();
  testAllPop();

  return 0;
}


uint32_t doubleFloatAsm(uint32_t in) {
  asm(""
      : "=r"(in) // output
      : "r"(in)  // input
      :          // clobbered registers
      );
  return in;
}

int popCountAsm(uint32_t in) {
  int pop = 0;
  asm(""
      : "=r"(pop) // output
      : "r"(in)   // input
      :           // clobbered registers
      );
  return pop;
}
